from sqlalchemy import create_engine
from sqlalchemy import Column, Integer, String, ForeignKey
from sqlalchemy.orm import sessionmaker, relationship
from sqlalchemy.ext.declarative import declarative_base

Base = declarative_base()
engine = create_engine('sqlite:///:memory:')
Session = sessionmaker(bind=engine)


class Toaster(Base):

    __tablename__ = "toasters"

    id = Column(Integer, primary_key=True)
    name = Column(String)
    color_id = Column(String, ForeignKey('colors.name'))

    color = relationship("Color", backref='toasters')


class Color(Base):

    __tablename__ = "colors"

    id = Column(Integer, primary_key=True)
    name = Column(String)



from charlatan import testing
from charlatan import FixturesManager
from charlatan.tests.fixtures.models import Session, Base, engine
from charlatan.tests.fixtures.models import Toaster, Color


class TestSqlalchemyFixtures(testing.TestCase):

    def setUp(self):
        self.session = Session()
        self.manager = FixturesManager(db_session=self.session)
        self.manager.load("./charlatan/tests/data/relationships.yaml")

        Base.metadata.create_all(engine)

    def tearDown(self):
        Base.metadata.drop_all(engine)
        self.session.close()

    def test_double_install(self):
        """Verify that there's no double install."""
        self.manager.install_fixture("model")
        self.manager.install_fixture("color")

        self.assertEqual(self.session.query(Toaster).count(), 1)
        self.assertEqual(self.session.query(Color).count(), 1)

    def test_getting_from_database(self):
        """Verify that we can get from the database."""
        installed = Toaster(id=1)
        self.session.add(installed)
        self.session.commit()

        toaster = self.manager.install_fixture("from_database")
        self.assertEqual(toaster.id, 1)

    def test_installing_collection(self):
        """Verify that a collection of fixtures is in the database."""
        self.manager.install_fixture("model_list")

        self.assertEqual(self.session.query(Toaster).count(), 2)

    def test_inheritance_and_relationship(self):
        """Verify that inheritance works with relationships."""
        model, model_1 = self.manager.install_fixtures(('model', 'model_1'))

        self.assertTrue(isinstance(model.color, Color))
        self.assertTrue(isinstance(model_1.color, Color))

    def test_explicit_foreign_key(self):
        """Verify that we can get a db-computed foreign key explicitely."""
        model = self.manager.install_fixture('model_with_explicit_fk')
        assert model.color_id is not None

    def test_uninstall_deletes_fixtures(self):
        """Verify uninstalling a fixture drops it from the database."""
        self.manager.install_fixture("color")

        # sanity check
        self.assertEqual(self.session.query(Color).count(), 1)

        self.manager.uninstall_fixture("color")

        self.assertEqual(self.session.query(Color).count(), 0)





class Toaster(object):

    def __init__(self, color, slots=2, content=None):
        self.color = color
        self.slots = slots
        self.content = content

    def __repr__(self):
        return "<Toaster '%s'>" % self.color


class User(object):

    def __init__(self, toasters):
        self.toasters = toasters



from __future__ import absolute_import
import datetime

import pytz
import yaml
from yaml.constructor import Constructor

from charlatan.utils import get_timedelta, datetime_to_epoch_timestamp

TIMEZONE_AWARE = True


class RelationshipToken(str):

    """Class used to mark relationships.

    This token is used to mark relationships found in YAML file, so that they
    can be processed later.
    """

    pass


class UnnamedRelationshipToken(dict):

    """Class used to mark unamed relationships.

    This token is used to mark relationships found in YAML file, so that they
    can be processed later.
    """

    pass


def configure_yaml():
    """Add some custom tags to the YAML constructor."""
    def now_constructor(loader, node):
        """Return a function that returns the current datetime."""
        delta = get_timedelta(loader.construct_scalar(node))

        def get_now():
            returned = datetime.datetime.utcnow()
            if TIMEZONE_AWARE:
                returned = returned.replace(tzinfo=pytz.utc)
            return returned + delta

        return get_now

    def now_naive_constructor(loader, node):
        """Return a function that returns the current datetime.

        The returned datetime is always a naive datetime (i.e. without
        timezone information).

        See the introduction in `datetime
        <https://docs.python.org/2/library/datetime.html>`_ for more
        information.
        """
        delta = get_timedelta(loader.construct_scalar(node))

        def get_now():
            return datetime.datetime.utcnow() + delta

        return get_now

    def epoch_now_constructor(loader, node):
        """Return a function that returns the current epoch."""
        delta = get_timedelta(loader.construct_scalar(node))

        def get_now():
            return datetime_to_epoch_timestamp(
                datetime.datetime.utcnow() + delta)

        return get_now

    def relationship_constructor(loader, node):
        """Create _RelationshipToken for `!rel` tags."""
        name = loader.construct_scalar(node)
        return RelationshipToken(name)

    yaml.add_constructor(u'!now', now_constructor)
    yaml.add_constructor(u'!now_naive', now_naive_constructor)
    yaml.add_constructor(u'!epoch_now', epoch_now_constructor)
    yaml.add_constructor(u'!rel', relationship_constructor)


def configure_output(use_unicode=False):
    """Configure output options of the values loaded by pyyaml.

    :param bool use_unicode: Use unicode constructor for loading strings
    """
    if use_unicode:
        yaml.add_constructor(
            u'tag:yaml.org,2002:str',
            Constructor.construct_python_unicode,
        )


def load_file(filename, use_unicode=False):
    """Load fixtures definition from file.

    :param str filename:
    """
    with open(filename) as f:
        content = f.read()

    if filename.endswith(".yaml"):
        # Load the custom YAML tags
        configure_yaml()
        configure_output(use_unicode=use_unicode)
        content = yaml.load(content)
    else:
        raise ValueError("Unsupported filetype: '%s'" % filename)

    return content



from __future__ import absolute_import
from sys import version_info
import datetime

from freezegun import freeze_time
from yaml import add_constructor
from yaml.constructor import SafeConstructor
import pytest
import pytz
import unittest

from charlatan import testing
from charlatan import file_format
from charlatan.utils import datetime_to_epoch_timestamp


def test_non_yaml_file():
    """Verify that we can't open a non-YAML file."""
    with pytest.raises(ValueError):
        file_format.load_file("./charlatan/tests/data/test.json")


@freeze_time("2014-12-31 11:00:00")
class TestFileFormat(testing.TestCase):

    def setUp(self):
        self.current_time = datetime.datetime.utcnow().replace(
            tzinfo=pytz.utc)
        self.yaml = file_format.load_file(
            './charlatan/tests/data/special_tags.yaml'
        )

    def test_now_tag(self):
        """Assert !now creates a current datetime."""
        self.assertEqual(self.current_time, self.yaml['current_time']())

    def test_now_naive_tag(self):
        """Assert !now_naive creates a current naive datetime."""
        assert datetime.datetime.utcnow() == self.yaml['current_naive_time']()

    def test_time_offsets(self):
        """Assert that !now +1d gives a day in the future."""
        tomorrow = self.current_time + datetime.timedelta(days=1)
        self.assertEqual(tomorrow, self.yaml['tomorrow']())

    def test_epoch_now_tag(self):
        """Assert !epoch_now gives integer time."""
        current_epoch_time = datetime_to_epoch_timestamp(self.current_time)
        self.assertEqual(current_epoch_time, self.yaml['current_epoch_time']())

    def test_epoch_now_tag_with_offset(self):
        """Assert !epoch_now accepts an offset."""
        tomorrow_datetime = self.current_time + datetime.timedelta(days=1)
        tomorrow = datetime_to_epoch_timestamp(tomorrow_datetime)

        self.assertEqual(tomorrow, self.yaml['tomorrow_epoch_time']())

    def test_rel_tag(self):
        """Assert !rel tag makes the value a relationship token."""
        self.assertIsInstance(
            self.yaml['relationship'], file_format.RelationshipToken)


class TestUnicodeLoad(testing.TestCase):

    def setUp(self):
        # preserve the original constructor for strings
        self.str_constructor = SafeConstructor.yaml_constructors[
            u'tag:yaml.org,2002:str'
        ]
        self.yaml = file_format.load_file(
            './charlatan/tests/data/strings.yaml',
            use_unicode=True,
        )

    def tearDown(self):
        # reset the constructor
        add_constructor(u'tag:yaml.org,2002:str', self.str_constructor)

    @unittest.skipIf(version_info[0] == 3, 'Unicode is undefined in Python 3')
    def test_strings_are_unicode(self):
        """Assert all strings are loaded as unicode."""
        for key, val in self.yaml.items():
            self.assertTrue(isinstance(key, unicode))  # noqa
            self.assertTrue(isinstance(val, unicode))  # noqa


class TestStringLoad(testing.TestCase):

    def setUp(self):
        self.yaml = file_format.load_file(
            './charlatan/tests/data/strings.yaml',
        )

    @unittest.skipIf(version_info[0] == 3, 'Iteration has changed in Python 3')
    def test_strings_are_strings(self):
        """Assert all strings are loaded as strings."""
        for key, val in self.yaml.items():
            self.assertTrue(isinstance(key, str))
            self.assertTrue(isinstance(val, str))

    @unittest.skipIf(version_info[0] == 2, 'Iteration has changed in Python 3')
    def test_strings_are_strings_python3(self):
        """Assert all strings are loaded as strings."""
        for key, val in list(self.yaml.items()):
            self.assertTrue(isinstance(key, str))
            self.assertTrue(isinstance(val, str))



from charlatan import _compat, utils
from charlatan.fixture import Inheritable


def _sorted_iteritems(dct):
    """Iterate over the items in the dict in a deterministic fashion."""
    for k, v in sorted(_compat.iteritems(dct)):
        yield k, v


class FixtureCollection(Inheritable):

    """A FixtureCollection holds Fixture objects."""

    def __init__(self, key, fixture_manager,
                 model=None,
                 models_package=None,
                 fields=None,
                 post_creation=None,
                 inherit_from=None,
                 depend_on=None,
                 fixtures=None):
        super(FixtureCollection, self).__init__()
        self.key = key
        self.fixture_manager = fixture_manager
        self.fixtures = fixtures or self.container()

        self.key = key

        self.inherit_from = inherit_from
        self._has_updated_from_parent = False

        # Stuff that can be inherited.
        self.fields = fields or {}
        self.model_name = model
        self.models_package = models_package
        self.post_creation = post_creation or {}
        self.depend_on = depend_on

    def __repr__(self):
        return "<%s '%s'>" % (self.__class__.__name__, self.key)

    def __iter__(self):
        return self.iterator(self.fixtures)

    def get_instance(self, path=None, overrides=None, builder=None):
        """Get an instance.

        :param str path:
        :param dict overrides:
        :param func builder:
        """
        if not path:
            return self.get_all_instances(overrides=overrides, builder=builder)

        remaining_path = ''
        if isinstance(path, _compat.string_types):
            path = path.split(".")
            first_level = path[0]
            remaining_path = ".".join(path[1:])

        # First try to get the fixture from the cache
        instance = self.fixture_manager.cache.get(first_level)
        if (not overrides
                and instance
                and not isinstance(instance, FixtureCollection)):
            if not remaining_path:
                return instance
            return utils.richgetter(instance, remaining_path)

        # Or just get it
        fixture = self.get(first_level)
        # Then we ask it to return an instance.
        return fixture.get_instance(path=remaining_path,
                                    overrides=overrides,
                                    builder=builder,
                                    )

    def get_all_instances(self, overrides=None, builder=None):
        """Get all instances.

        :param dict overrides:
        :param func builder:

        .. deprecated:: 0.4.0
            Removed format argument.
        """
        returned = []
        for name, fixture in self:
            instance = fixture.get_instance(overrides=overrides,
                                            builder=builder)
            returned.append((name, instance))

        if self.container is dict:
            return dict(returned)
        elif self.container is list:
            return list(map(lambda f: f[1], returned))
        else:
            raise ValueError('Unknown container')

    def extract_relationships(self):
        # Just proxy to fixtures in this collection.
        for _, fixture in self:
            for r in fixture.extract_relationships():
                yield r


class DictFixtureCollection(FixtureCollection):
    iterator = staticmethod(_sorted_iteritems)
    container = dict

    def add(self, name, fixture):
        self.fixtures[str(name)] = fixture

    def get(self, path):
        """Return a single fixture.

        :param str path:
        """
        if path not in self.fixtures:
            raise KeyError("No such fixtures: '%s'" % path)

        return self.fixtures[path]


class ListFixtureCollection(FixtureCollection):
    iterator = enumerate
    container = list

    def add(self, _, fixture):
        self.fixtures.append(fixture)

    def get(self, path):
        """Return a single fixture.

        :param str path:
        """
        return self.fixtures[int(path)]



import pytest

from charlatan import FixturesManager


def get_collection(collection):
    """Return FixtureCollection.

    :param str collection: name of collection to import
    """
    manager = FixturesManager()
    manager.load("docs/examples/collection.yaml")
    return manager.collection.get(collection)


@pytest.fixture(scope="module")
def collection():
    """Return FixtureCollection."""
    return get_collection("toasters")


def test_repr(collection):
    """Verify the repr."""
    assert repr(collection) == "<DictFixtureCollection 'toasters'>"


def test_cant_get_missing_fixture(collection):
    """Verify that we can't get a missing fixture."""
    with pytest.raises(KeyError):
        collection.get("missing")



from __future__ import absolute_import

from charlatan import testing
from charlatan.utils import richgetter


class TestRichGetter(testing.TestCase):
    def test_resolves_objects(self):
        obj = Namespace(key=Namespace(value='value'))

        result = richgetter(obj, 'key.value')

        self.assertEquals(result, 'value')

    def test_resolves_mappings(self):
        obj = dict(key=dict(value='value'))

        result = richgetter(obj, 'key.value')

        self.assertEquals(result, 'value')

    def test_resolves_sequences(self):
        obj = [['value']]

        result = richgetter(obj, '0.0')

        self.assertEquals(result, 'value')

    def test_resolves_mixed_containers(self):
        obj = Namespace(key=dict(values=['value']))

        result = richgetter(obj, 'key.values.0')

        self.assertEquals(result, 'value')


class Namespace(object):
    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)



"""Utilities for writing code that runs on Python 2 and 3."""

# Copyright (c) 2010-2013 Benjamin Peterson
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.
# IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY
# CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT,
# TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
# SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import sys

# Useful for very coarse version differentiation.
PY2 = sys.version_info[0] == 2
PY3 = sys.version_info[0] == 3

if PY3:
    string_types = str,

else:
    string_types = basestring,  # noqa


if PY3:
    _iteritems = "items"
    _itervalues = "values"

else:
    _iteritems = "iteritems"
    _itervalues = "itervalues"


def iteritems(d, **kw):
    """Return an iterator over the (key, value) pairs of a dictionary."""
    return iter(getattr(d, _iteritems)(**kw))


def itervalues(d, **kw):
    """Return an iterator over the values of a dictionary."""
    return iter(getattr(d, _itervalues)(**kw))



import calendar
import datetime
import functools
import itertools
import operator
import re
import collections

from charlatan import _compat

VALID_SIGNS = frozenset(['-', '+'])


def get_timedelta(delta):
    """Return timedelta from string.

    :param str delta:

    :rtype: :py:class:`datetime.timedelta` instance

    >>> get_timedelta("")
    datetime.timedelta(0)
    >>> get_timedelta("+1h")
    datetime.timedelta(0, 3600)
    >>> get_timedelta("+10h")
    datetime.timedelta(0, 36000)
    >>> get_timedelta("-10d")
    datetime.timedelta(-10)
    >>> get_timedelta("+1m")
    datetime.timedelta(30)
    >>> get_timedelta("-1y")
    datetime.timedelta(-365)
    >>> get_timedelta("+10d2h")
    datetime.timedelta(10, 7200)
    >>> get_timedelta("-10d2h")
    datetime.timedelta(-11, 79200)
    >>> get_timedelta("-21y2m1d24h")
    datetime.timedelta(-7727)
    >>> get_timedelta("+5M")
    datetime.timedelta(0, 300)

    """
    if not delta:
        return datetime.timedelta()
    sign = delta[0]
    assert sign in VALID_SIGNS
    delta = delta[1:]
    timedelta_kwargs = {}

    # re.split returns empty strings if the capturing groups matches the
    # start and the end of string.
    for part in filter(lambda p: p, re.split(r"(\d+\w)", delta)):
        amount, unit = re.findall(r"(\d+)([ymdhMs])", part)[0]
        units = {
            "s": "seconds",
            "M": "minutes",
            "h": "hours",
            "d": "days",
            "m": "months",
            "y": "years"
        }
        timedelta_kwargs[units[unit]] = int(amount)

    delta = extended_timedelta(**timedelta_kwargs)
    if sign == '-':
        return -delta
    else:
        return delta


def extended_timedelta(**kwargs):
    """Return a :py:class:`timedelta` object based on the arguments.

    :param integer years:
    :param integer months:
    :param integer days:
    :rtype: :py:class:`timedelta` instance

    Since :py:class:`timedelta`'s largest unit are days, :py:class:`timedelta`
    objects cannot be created with a number of months or years as an argument.
    This function lets you create :py:class:`timedelta` objects based on a
    number of days, months and years.

    >>> extended_timedelta(months=1)
    datetime.timedelta(30)
    >>> extended_timedelta(years=1)
    datetime.timedelta(365)
    >>> extended_timedelta(days=1, months=1, years=1)
    datetime.timedelta(396)
    >>> extended_timedelta(hours=1)
    datetime.timedelta(0, 3600)
    """
    number_of_days = {
        "days": 1,
        "months": 30,
        "years": 365}

    days = []
    kwargs_copy = kwargs.copy()  # So that we can remove values from kwargs
    for k in kwargs_copy:
        if k in number_of_days:
            days.append([number_of_days[k], kwargs.pop(k)])

    if days:
        add = int(functools.reduce(
            operator.add, itertools.starmap(operator.mul, days)))
        kwargs["days"] = kwargs.setdefault("days", 0) + add

    return datetime.timedelta(**kwargs)


def datetime_to_epoch_timestamp(a_datetime):
    """Return the epoch timestamp for the given datetime.

    :param datetime a_datetime: The datetime to translate
    :rtype: float

    >>> a_datetime = datetime.datetime(2013, 11, 21, 1, 33, 11, 160611)
    >>> datetime_to_epoch_timestamp(a_datetime)
    1384997591.160611
    """
    return (calendar.timegm(a_datetime.utctimetuple()) + a_datetime.microsecond / 1000000.0)  # noqa


# TODO: does not copy the function signature
# see http://stackoverflow.com/questions/2982974/copy-call-signature-to-decorator  # noqa


def copy_docstring_from(klass):
    """Copy docstring from another class, using the same function name."""
    def wrapper(func):
        func.__doc__ = getattr(klass, func.__name__).__doc__

        @functools.wraps(func)
        def wrapped(*args, **kwargs):
            return func(*args, **kwargs)

        return wrapped

    return wrapper


def safe_iteritems(items):
    """Safely iterate over a dict or a list."""
    # For dictionaries, iterate over key, value and for lists iterate over
    # index, item
    if hasattr(items, 'items'):
        return _compat.iteritems(items)
    else:
        return enumerate(items)


def is_sqlalchemy_model(instance):
    """Return True if instance is an SQLAlchemy model instance."""
    from sqlalchemy.orm.util import class_mapper
    from sqlalchemy.orm.exc import UnmappedClassError

    try:
        class_mapper(instance.__class__)

    except UnmappedClassError:
        return False

    else:
        return True


def richgetter(obj, path):
    """Return a attrgetter + item getter."""
    for name in path.split("."):
        if isinstance(obj, collections.Mapping):
            obj = obj[name]
        elif isinstance(obj, collections.Sequence):
            obj = obj[int(name)]  # force int type for list indexes
        else:
            obj = getattr(obj, name)

    return obj



from __future__ import print_function
from glob import glob
from itertools import chain
import functools
import os

from charlatan import _compat
from charlatan import builder
from charlatan.depgraph import DepGraph
from charlatan.file_format import load_file
from charlatan.fixture import Fixture
from charlatan import fixture_collection

ALLOWED_HOOKS = ("before_save", "after_save", "before_install",
                 "after_install")
ROOT_COLLECTION = "root"


def make_list(obj):
    """Return list of objects if necessary."""
    if isinstance(obj, _compat.string_types):
        return (obj, )
    return obj


class FixturesManager(object):

    """
    Manage Fixture objects.

    :param Session db_session: sqlalchemy Session object
    :param bool use_unicode:
    :param func get_builder:
    :param func delete_builder:

    .. versionadded:: 0.4.0
        ``get_builder`` and ``delete_builder`` arguments were added.

    .. deprecated:: 0.4.0
        ``delete_instance``, ``save_instance`` methods were deleted in favor
        of using builders.

    .. versionadded:: 0.3.0
        ``db_session`` argument was added.
    """

    DictFixtureCollection = fixture_collection.DictFixtureCollection
    ListFixtureCollection = fixture_collection.ListFixtureCollection

    default_get_builder = builder.InstantiateAndSave()
    default_delete_builder = builder.DeleteAndCommit()

    def __init__(self, db_session=None, use_unicode=False,
                 get_builder=None, delete_builder=None,
                 ):
        self.hooks = {}
        self.session = db_session
        self.installed_keys = []
        self.use_unicode = use_unicode
        self.get_builder = get_builder or self.default_get_builder
        self.delete_builder = delete_builder or self.default_delete_builder
        self.filenames = []
        self.collection = self.DictFixtureCollection(
            ROOT_COLLECTION,
            fixture_manager=self,
        )

    def load(self, filenames, models_package=""):
        """Pre-load the fixtures. Does not install anything.

        :param list_or_str filename: file or list of files that holds the
                                     fixture data
        :param str models_package: package holding the models definition

        .. deprecated:: 0.3.0
            ``db_session`` argument was removed and put in the object's
            constructor arguments.

        .. versionchanged:: 0.3.7
            ``filename`` argument was changed to ``filenames``, which can be
            list or string.

        """
        self.filenames.append(filenames)

        self.depgraph = self._load_fixtures(filenames,
                                            models_package=models_package)
        self.clean_cache()

    def _get_namespace_from_filename(self, filename):
        """Get a collection namespace from a fixtures filename.

        :param str filename: filename to extract namespace from
        """
        segments = os.path.basename(filename).split(".")
        if len(segments) > 2:
            raise ValueError("Fixtures filename stem may not contain periods")

        return segments[0]

    def _load_fixtures(self, filenames, models_package=''):
        """Pre-load the fixtures.

        :param list or str filenames: files that hold the fixture data
        :param str models_package:
        """
        if isinstance(filenames, _compat.string_types):
            globbed_filenames = glob(filenames)
        else:
            globbed_filenames = list(
                chain.from_iterable(glob(f) for f in filenames)
            )

        if not globbed_filenames:
            raise IOError('File "%s" not found' % filenames)

        if len(globbed_filenames) == 1:
            content = load_file(filenames, self.use_unicode)
        else:
            content = {}

            for filename in globbed_filenames:
                namespace = self._get_namespace_from_filename(filename)
                content[namespace] = {
                    "objects": load_file(filename, self.use_unicode)
                }

        if content:
            for k, v in _compat.iteritems(content):

                if "objects" in v:
                    # It's a collection of fictures.
                    collection = self._handle_collection(
                        namespace=k,
                        definition=v,
                        objects=v["objects"],
                        models_package=models_package,
                    )
                    self.collection.add(k, collection)

                # Named fixtures
                else:
                    if "id" in v:
                        # Renaming id because it's a Python builtin function
                        v["id_"] = v["id"]
                        del v["id"]

                    fixture = Fixture(
                        key=k,
                        fixture_manager=self,
                        models_package=models_package,
                        **v)
                    self.collection.add(k, fixture)

        graph = self._check_cycle(self.collection)
        return graph

    def _check_cycle(self, collection):
        """Raise an exception if there's a relationship cycle."""
        d = DepGraph()
        for _, fixture in collection:
            for dependency, _ in fixture.extract_relationships():
                d.add_edge(dependency, fixture.key)

        # This does nothing except raise an error if there's a cycle
        d.topo_sort()
        return d

    def _handle_collection(self, namespace, definition, objects,
                           models_package=''):
        """Handle a collection of fixtures.

        :param dict definition: definition of the collection
        :param dict_or_list objects: fixtures in the collection
        :param str models_package:

        """
        if isinstance(objects, list):
            klass = self.ListFixtureCollection
        else:
            klass = self.DictFixtureCollection

        collection = klass(
            key=namespace,
            fixture_manager=self,
            model=definition.get('model'),
            models_package=definition.get('models_package'),
            fields=definition.get('fields'),
            post_creation=definition.get('post_creation'),
            inherit_from=definition.get('inherit_from'),
            depend_on=definition.get('depend_on'),
        )

        for name, new_fields in collection.iterator(objects):
            qualified_name = "%s.%s" % (namespace, name)

            if "objects" in new_fields:
                # A nested collection, either because we're dealing with a file
                # collection or a sub-collection.
                fixture = self._handle_collection(
                    namespace=qualified_name,
                    definition=new_fields,
                    objects=new_fields["objects"]
                )
            else:
                model = new_fields.pop("model", None)
                # In the case of a file collection we'll be dealing with
                # PyYAML's output from that file, which means that individual
                # fixtures in this collection have the "fields" field.
                fields = new_fields.pop("fields", new_fields)
                inherit_from = namespace if model is None else None

                fixture = Fixture(
                    key=qualified_name,
                    fixture_manager=self,
                    # Automatically inherit from the collection
                    inherit_from=inherit_from,
                    fields=fields,
                    model=model,
                    models_package=models_package,
                    # The rest (default fields, etc.) is
                    # automatically inherited from the collection.
                )
            collection.add(name, fixture)

        return collection

    def clean_cache(self):
        """Clean the cache."""
        self.cache = {}
        self.installed_keys = []

    def delete_fixture(self, fixture_key, builder=None):
        """Delete a fixture instance.

        :param str fixture_key:
        :param func builder:

        Before and after the process, the :func:`before_delete` and
        :func:`after_delete` hook are run.

        .. versionadded:: 0.4.0
            ``builder`` argument was added.

        .. deprecated:: 0.4.0
            ``delete_instance`` method renamed to ``delete_fixture`` for
            consistency reason.
        """
        builder = builder or self.delete_builder
        self.get_hook("before_delete")(fixture_key)

        instance = self.cache.get(fixture_key)
        if instance:
            self.cache.pop(fixture_key, None)
            self.installed_keys.remove(fixture_key)
            builder(self, instance)

        self.get_hook("after_delete")(fixture_key)

    def install_fixture(self, fixture_key, overrides=None):
        """Install a fixture.

        :param str fixture_key:
        :param dict overrides: override fields

        :rtype: :data:`fixture_instance`

        .. deprecated:: 0.4.0
            ``do_not_save`` argument was removed.
            ``attrs`` argument renamed ``overrides``.

        .. deprecated:: 0.3.7
            ``include_relationships`` argument was removed.

        """
        builder = functools.partial(self.get_builder,
                                    save=True,
                                    session=self.session)
        self.get_hook("before_install")()

        try:
            instance = self.get_fixture(fixture_key, overrides=overrides,
                                        builder=builder)
        except Exception as exc:
            self.get_hook("after_install")(exc)
            raise

        else:
            self.get_hook("after_install")(None)
            return instance

    def install_fixtures(self, fixture_keys):
        """Install a list of fixtures.

        :param fixture_keys: fixtures to be installed
        :type fixture_keys: str or list of strs
        :rtype: list of :data:`fixture_instance`

        .. deprecated:: 0.4.0
            ``do_not_save`` argument was removed.

        .. deprecated:: 0.3.7
            ``include_relationships`` argument was removed.

        """
        instances = []
        for f in make_list(fixture_keys):
            instances.append(self.install_fixture(f))
        return instances

    def install_all_fixtures(self):
        """Install all fixtures.

        :rtype: list of :data:`fixture_instance`

        .. deprecated:: 0.4.0
            ``do_not_save`` argument was removed.

        .. deprecated:: 0.3.7
            ``include_relationships`` argument was removed.

        """
        return self.install_fixtures(self.keys())

    def uninstall_fixture(self, fixture_key):
        """Uninstall a fixture.

        :param str fixture_key:
        :rtype: ``None``

        .. deprecated:: 0.4.0
            ``do_not_delete`` argument was removed. This function does not
            return anything.
        """
        builder = functools.partial(self.delete_builder,
                                    commit=True,
                                    session=self.session)
        return self.delete_fixture(fixture_key, builder)

    def uninstall_fixtures(self, fixture_keys):
        """Uninstall a list of installed fixtures.

        :param fixture_keys: fixtures to be uninstalled
        :type fixture_keys: str or list of strs
        :rtype: ``None``

        .. deprecated:: 0.4.0
            ``do_not_delete`` argument was removed. This function does not
            return anything.
        """
        for fixture_key in make_list(fixture_keys):
            self.uninstall_fixture(fixture_key)

    def uninstall_all_fixtures(self):
        """Uninstall all installed fixtures.

        :rtype: ``None``

        .. deprecated:: 0.4.0
            ``do_not_delete`` argument was removed. This function does not
            return anything.
        """
        installed_fixtures = list(self.installed_keys)
        installed_fixtures.reverse()
        self.uninstall_fixtures(installed_fixtures)

    def keys(self):
        """Return all fixture keys."""
        return self.collection.fixtures.keys()

    def get_fixture(self, fixture_key, overrides=None, builder=None):
        """Return a fixture instance (but do not save it).

        :param str fixture_key:
        :param dict overrides: override fields
        :param func builder: build builder.
        :rtype: instantiated but unsaved fixture

        .. versionadded:: 0.4.0
            ``builder`` argument was added.
            ``attrs`` argument renamed ``overrides``.

        .. deprecated:: 0.4.0
            ``do_not_save`` argument was removed.

        .. deprecated:: 0.3.7
            ``include_relationships`` argument was removed.
        """
        builder = builder or self.get_builder
        # initialize all parents in topological order
        parents = []
        for fixture in self.depgraph.ancestors_of(fixture_key):
            parents.append(self.get_fixture(fixture, builder=builder))

        # Fixture are cached so that setting up relationships is not too
        # expensive. We don't get the cached version if overrides are
        # overriden.
        returned = None
        if not overrides:
            returned = self.cache.get(fixture_key)

        if not returned:
            returned = self.collection.get_instance(
                fixture_key, overrides=overrides, builder=builder)

            self.cache[fixture_key] = returned
            self.installed_keys.append(fixture_key)

        return returned

    def get_fixtures(self, fixture_keys, builder=None):
        """Get fixtures from iterable.

        :param iterable fixture_keys:
        :rtype: list of instantiated but unsaved fixtures

        .. versionadded:: 0.4.0
            ``builder`` argument was added.

        .. deprecated:: 0.3.7
            ``include_relationships`` argument was removed.
        """
        builder = builder or self.get_builder
        fixtures = []
        for f in fixture_keys:
            fixtures.append(self.get_fixture(f))
        return fixtures

    def get_all_fixtures(self, builder=None):
        """Get all fixtures.

        :param iterable fixture_keys:
        :rtype: list of instantiated but unsaved fixtures

        .. versionadded:: 0.4.0
            ``builder`` argument was added.

        .. deprecated:: 0.3.7
            ``include_relationships`` argument was removed.
        """
        builder = builder or self.get_builder
        return self.get_fixtures(self.keys(), builder=builder)

    def get_hook(self, hook_name):
        """Return a hook.

        :param str hook_name: e.g. ``before_delete``.
        """
        if hook_name in self.hooks:
            return self.hooks[hook_name]

        return lambda *args: None

    def set_hook(self, hookname, func):
        """Add a hook.

        :param str hookname:
        :param function func:
        """
        if hookname not in ALLOWED_HOOKS:
            raise KeyError("'%s' is not an allowed hook." % hookname)

        self.hooks[hookname] = func



from charlatan import FixturesManager
from charlatan.builder import Builder


class Toaster(object):

    def __init__(self, attrs):
        self.slots = attrs['slots']


class DictBuilder(Builder):

    def __call__(self, fixtures, klass, params, **kwargs):
        # A "normal" object would be instantiated this way:
        # return klass(**params)

        # Yet schematics object expect a dict of attributes as only argument.
        # So we'll do:
        return klass(params)


def test_custom_builder():
    manager = FixturesManager(get_builder=DictBuilder())
    manager.load('./charlatan/tests/example/data/custom_builder.yaml')
    assert manager.get_fixture('toaster').slots == 3



from charlatan.utils import is_sqlalchemy_model


class Builder(object):

    def __call__(self, fixtures, klass, params, **kwargs):
        """Build a fixture.

        :param FixturesManager fixtures:
        :param klass: the fixture's class (``model`` in the definition file)
        :param params: the fixture's params (``fields`` in the definition
            file)
        :param dict kwargs:

        ``kwargs`` allows passing arguments to the builder to change its
        behavior.
        """
        raise NotImplementedError


class InstantiateAndSave(Builder):

    def __call__(self, fixtures, klass, params, **kwargs):
        """Save a fixture instance.

        If it's a SQLAlchemy model, it will be added to the session and
        the session will be committed.

        Otherwise, a :meth:`save` method will be run if the instance has
        one. If it does not have one, nothing will happen.

        Before and after the process, the :func:`before_save` and
        :func:`after_save` hook are run.

        """
        session = kwargs.get('session')
        save = kwargs.get('save')

        instance = self.instantiate(klass, params)
        if save:
            self.save(instance, fixtures, session)
        return instance

    def instantiate(self, klass, params):
        """Return instantiated instance."""
        try:
            return klass(**params)
        except TypeError as exc:
            raise TypeError("Error while trying to build %r "
                            "with %r: %s" % (klass, params, exc))

    def save(self, instance, fixtures, session):
        """Save instance."""
        fixtures.get_hook("before_save")(instance)

        if session and is_sqlalchemy_model(instance):
            session.add(instance)
            session.commit()

        else:
            getattr(instance, "save", lambda: None)()

        fixtures.get_hook("after_save")(instance)


class DeleteAndCommit(Builder):

    def __call__(self, fixtures, instance, **kwargs):
        session = kwargs.get('session')
        commit = kwargs.get('commit')

        fixtures.get_hook("before_uninstall")()
        try:
            if commit:
                self.delete(instance, session)
            else:
                try:
                    getattr(instance, "delete_instance")()
                except AttributeError:
                    getattr(instance, "delete", lambda: None)()

        except Exception as exc:
            fixtures.get_hook("after_uninstall")(exc)
            raise

        else:
            fixtures.get_hook("after_uninstall")(None)

    def delete(self, instance, session):
        """Delete instance."""
        if session and is_sqlalchemy_model(instance):
            session.delete(instance)
            session.commit()



from charlatan import testing
from charlatan import FixturesManager
from charlatan.tests.fixtures.models import Session, Base, engine
from charlatan.tests.fixtures.models import Toaster

session = Session()
manager = FixturesManager(db_session=session)
manager.load("./charlatan/tests/example/data/sqlalchemy.yaml")


class TestSqlalchemyFixtures(testing.TestCase):

    def setUp(self):
        self.manager = manager

        # There's a lot of different patterns to setup and teardown the
        # database. This is the simplest possibility.
        Base.metadata.create_all(engine)

    def tearDown(self):
        Base.metadata.drop_all(engine)
        session.close()

    def test_double_install(self):
        """Verify that there's no double install."""
        self.manager.install_fixture('toaster')

        toaster = session.query(Toaster).one()
        assert toaster.color.name == 'red'



from __future__ import absolute_import
import operator as op

from charlatan import testing
from charlatan import FixturesManager


class TestListOfFixtures(testing.TestCase):

    def setUp(self):
        self.fm = FixturesManager()
        self.fm.load('./charlatan/tests/data/lists.yaml')

    def test_get_list_by_name(self):
        """Verify that lists of fixtures returns lists."""
        fixtures = self.fm.install_fixture('fixture_list')
        self.assertIsInstance(fixtures, list)

    def test_one_to_many_relationship(self):
        """Verify that relations to lists of fixtures work."""
        fixture = self.fm.install_fixture('related_fixture')
        self.assertEqual(
            fixture['elements'],
            self.fm.install_fixture('fixture_list')
        )

    def test_override(self):
        """Verify that we can override attributes on a list of fixtures."""
        fixtures = self.fm.install_fixture('fixture_list',
                                           overrides={"field1": 12})
        assert list(map(op.itemgetter('field1'), fixtures)) == [12, 12]



import copy
import importlib

from charlatan import _compat
from charlatan.file_format import RelationshipToken
from charlatan.utils import safe_iteritems, richgetter

CAN_BE_INHERITED = frozenset(
    ["model_name", "models_package", "fields", "post_creation", "depend_on"])


def get_class(module, klass):
    """Return a class object.

    :param str module: module path
    :param str klass: class name
    """
    try:
        object_module = importlib.import_module(module)
        cls = getattr(object_module, klass)
    except (ImportError, TypeError) as e:
        raise ImportError("Unable to import %s:%s. %r" % (module, klass, e))

    return cls


class Inheritable(object):

    def __init__(self, *args, **kwargs):
        # This is to make sure we don't redo the inheritance twice, for
        # performance reason.
        self._has_inherited_from_parent = False
        self.inherit_from = None
        self.fixture_manager = None

    def inherit_from_parent(self):
        """Inherit the attributes from parent, modifying itself."""
        if self._has_inherited_from_parent or not self.inherit_from:
            # Nothing to do
            return

        for name, value in self.get_parent_values():
            setattr(self, name, value)

    def get_parent_values(self):
        """Return parent values."""
        parent = self.fixture_manager.collection.get(self.inherit_from)
        # Recursive to make sure everything is updated.
        parent.inherit_from_parent()

        for key in CAN_BE_INHERITED:
            children_value = getattr(self, key)
            parent_value = getattr(parent, key)
            new_value = None
            if not children_value:
                # If I don't have a value, then we try from the parent
                # no matter what.
                new_value = copy.deepcopy(parent_value)

            elif isinstance(children_value, _compat.string_types):
                # The children value is something, then it takes
                # precedence no matter what.
                continue

            elif hasattr(children_value, "update"):
                # If it's a dict, then we try inheriting from the
                # parent.
                new_value = copy.deepcopy(parent_value)
                new_value.update(children_value)

            if new_value:
                yield key, new_value


class Fixture(Inheritable):

    """Represent a fixture that can be installed."""

    def __init__(self, key, fixture_manager,
                 model=None, fields=None,
                 inherit_from=None,
                 post_creation=None, id_=None,
                 models_package='',
                 depend_on=frozenset()):
        """Create a Fixture object.

        :param str model: model used to instantiate the fixture, e.g.
            "yourlib.toaster:Toaster". If empty, the fields will be used as is.
        :param str models_package: default models package for relative imports
        :param dict fields: args to be provided when instantiating the fixture
        :param fixture_manager: FixturesManager creating the fixture
        :param dict post_creation: assignment to be done after instantiation
        :param str inherit_from: model to inherit from
        :param list depend_on: A list of relationships to depend on

        .. versionadded:: 0.4.0
            ``models_package`` argument added.

        """
        super(Fixture, self).__init__()

        if id_ and fields:
            raise TypeError(
                "Cannot provide both id and fields to create fixture.")

        self.key = key
        self.fixture_manager = fixture_manager

        self.database_id = id_
        self.inherit_from = inherit_from

        # Stuff that can be inherited.
        self.model_name = model
        self.models_package = models_package
        self.fields = fields or {}
        self.post_creation = post_creation or {}
        self.depend_on = depend_on

    def __repr__(self):
        return "<Fixture '%s'>" % self.key

    def get_instance(self, path=None, overrides=None, builder=None):
        """Instantiate the fixture using the model and return the instance.

        :param str path: remaining path to return
        :param dict overrides: overriding fields
        :param func builder: function that is used to get the fixture

        .. deprecated:: 0.4.0
            ``fields`` argument renamed ``overrides``.

        .. versionadded:: 0.4.0
            ``builder`` argument added.

        .. deprecated:: 0.3.7
            ``include_relationships`` argument removed.

        """
        self.inherit_from_parent()  # Does the modification in place.

        if self.database_id:
            object_class = self.get_class()
            # No need to create a new object, just get it from the db
            instance = (
                self.fixture_manager.session.query(object_class)
                .get(self.database_id)
            )

        else:
            # We need to do a copy since we're modifying them.
            params = copy.deepcopy(self.fields)
            if overrides:
                params.update(overrides)

            for key, value in safe_iteritems(params):
                if callable(value):
                    params[key] = value()

            # Get the class
            object_class = self.get_class()

            # Does not return anything, does the modification in place (in
            # fields).
            self._process_relationships(params)

            if object_class:
                instance = builder(self.fixture_manager, object_class, params)
            else:
                # Return the fields as is. This allows to enter dicts
                # and lists directly.
                instance = params

        # Do any extra assignment
        for attr, value in self.post_creation.items():
            if isinstance(value, RelationshipToken):
                value = self.get_relationship(value)

            setattr(instance, attr, value)

        if path:
            return richgetter(instance, path)
        else:
            return instance

    def get_class(self):
        """Return class object for this instance."""
        if not self.model_name:
            return

        # Relative path, e.g. ".toaster:Toaster"
        if ":" in self.model_name and self.model_name[0] == ".":
            module, klass = self.model_name.split(":")
            module = self.models_package + module
            return get_class(module, klass)

        # Absolute import, e.g. "yourlib.toaster:Toaster"
        if ":" in self.model_name:
            module, klass = self.model_name.split(":")
            return get_class(module, klass)

        # Class alone, e.g. "Toaster".
        # Trying to import from e.g.  yourlib.toaster:Toaster
        module = "{models_package}.{model}".format(
            models_package=self.models_package,
            model=self.model_name.lower())
        klass = self.model_name

        try:
            return get_class(module, klass)
        except ImportError:
            # Then try to import from yourlib:Toaster
            return get_class(self.models_package, klass)

    @staticmethod
    def extract_rel_name(name):
        """Return the relationship and attr from an argument to !rel."""
        rel_name = name  # e.g. toaster.color
        attr = None

        # TODO: we support only one level for now
        if "." in name:
            path = name.split(".")
            rel_name, attr = path[0], path[1]
        return rel_name, attr

    def extract_relationships(self):
        """Return all dependencies.

        :rtype generator:

        Yields ``(depends_on, attr_name)``.

        """
        # TODO: make this DRYer since it's mostly copied from
        # _process_relationships

        for dep in self.depend_on:
            yield dep, None

        for name, value in safe_iteritems(self.fields):
            # One to one relationship
            if isinstance(value, RelationshipToken):
                yield self.extract_rel_name(value)

            # One to many relationship
            elif isinstance(value, (tuple, list)):
                for i, nested_value in enumerate(value):
                    if isinstance(nested_value, RelationshipToken):
                        yield self.extract_rel_name(nested_value)

    def _process_field_relationships(self, field_value):
        """Create any relationship for a field if needed.

        :param mixed field_value: field value to be processed

        For each field that is a relationship or a list of relationships,
        instantiate those relationships and update the fields.

        Returns new field value and modifies lists in place.
        """
        # One to one relationship
        if isinstance(field_value, RelationshipToken):
            return self.get_relationship(field_value)

        # One to many relationship
        elif isinstance(field_value, (tuple, list)):
            for i, nested_value in enumerate(field_value):
                field_value[i] = self._process_field_relationships(
                    nested_value)

        return field_value

    def _process_relationships(self, fields):
        """Create any relationship if needed.

        :param dict fields: fields to be processed

        For each field that is a relationship or a list of relationships,
        instantiate those relationships and update the fields.

        Does not return anything, modify fields in place.
        """
        # For dictionaries, iterate over key, value and for lists iterate over
        # index, item
        if hasattr(fields, 'items'):
            field_iterator = _compat.iteritems(fields)
        else:
            field_iterator = enumerate(fields)

        for name, value in field_iterator:
            fields[name] = self._process_field_relationships(value)

    def get_relationship(self, name):
        """Get a relationship and its attribute if necessary."""
        # This function is needed so that this fixture can require other
        # fixtures. If a fixture requires another fixture, it
        # necessarily means that it needs to include other relationships
        # as well.
        return self.fixture_manager.get_fixture(name)



from __future__ import absolute_import
from datetime import datetime

import pytest
import pytz
from freezegun import freeze_time

from charlatan import testing
from charlatan import depgraph
from charlatan import FixturesManager


def test_overrides_and_in_cache():
    manager = FixturesManager()
    manager.load('./docs/examples/simple_fixtures.yaml')
    # Add it to the cache
    manager.install_fixture("toaster")
    toaster = manager.install_fixture("toaster", overrides={"color": "blue"})
    assert toaster.color == 'blue'


class TestFixturesManager(testing.TestCase):

    def test_load_two_files(self):
        """Verify we can load two files."""
        manager = FixturesManager()
        manager.load(
            './charlatan/tests/data/relationships_without_models.yaml')
        manager.load(
            './charlatan/tests/data/simple.yaml')
        assert 'simple_dict' in manager.keys()

    def test_load_empty_file(self):
        """Verify we can load a emtpy file."""
        manager = FixturesManager()
        manager.load('./charlatan/tests/data/empty.yaml')
        self.assertEqual(list(manager.keys()), [])

    def test_install_fixture(self):
        """install_fixture should return the fixture."""
        manager = FixturesManager()
        manager.load(
            './charlatan/tests/data/relationships_without_models.yaml')

        fixture = manager.install_fixture('simple_dict')

        self.assertEqual(fixture, {
            'field1': 'lolin',
            'field2': 2,
        })

    def test_get_all_fixtures(self):
        manager = FixturesManager()
        manager.load('./charlatan/tests/data/simple.yaml')
        assert len(manager.get_all_fixtures()) == 1

    def test_uninstall_all_fixtures(self):
        manager = FixturesManager()
        manager.load('./charlatan/tests/data/simple.yaml')
        assert len(manager.install_all_fixtures()) == 1
        manager.uninstall_all_fixtures()
        assert len(manager.installed_keys) == 0

    @freeze_time("2014-12-31 11:00:00")
    def test_install_fixture_with_now(self):
        """Verify that we can install a fixture with !now tag."""
        manager = FixturesManager()
        manager.load('./charlatan/tests/data/simple.yaml')
        fixture = manager.install_fixture('fixture')
        self.assertEqual(fixture,
                         {'now': datetime(2014, 12, 30, 11, 0,
                                          tzinfo=pytz.utc)})

    def test_install_fixture_override(self):
        """Verify that we can override a fixture field."""
        manager = FixturesManager()
        manager.load('./charlatan/tests/data/simple.yaml')
        fixture = manager.install_fixture('fixture', overrides={'now': None})
        self.assertEqual(fixture, {'now': None})

    def test_uninstall_fixture(self):
        manager = FixturesManager()
        manager.load(
            './charlatan/tests/data/relationships_without_models.yaml')

        manager.install_fixture('simple_dict')
        manager.uninstall_fixture('simple_dict')

        # verify we are forgiving with list inputs
        manager.install_fixtures('simple_dict')
        manager.uninstall_fixtures('simple_dict')

    def test_uninstall_non_installed_fixture(self):
        manager = FixturesManager()
        manager.load(
            './charlatan/tests/data/relationships_without_models.yaml')
        manager.uninstall_fixture('simple_dict')

    def test_dependency_parsing(self):
        fm = FixturesManager()
        fm.load(
            './charlatan/tests/data/dependencies.yaml'
        )
        assert fm.depgraph.has_edge_between('fixture1', 'fixture2')
        assert fm.depgraph.has_edge_between('fixture1', 'fixture3')
        assert fm.depgraph.has_edge_between('fixture4', 'fixture3')
        assert fm.depgraph.has_edge_between('fixture2', 'fixture4')

    def test_notices_cyclic_dependencies(self):
        fm = FixturesManager()
        self.assertRaises(
            depgraph.HasACycle,
            fm.load,
            './charlatan/tests/data/cyclic_dependencies.yaml'
        )

    def test_constructs_ancestors(self):
        fm = FixturesManager()
        fm.load(
            './charlatan/tests/data/dependencies.yaml'
        )
        assert not fm.cache
        # loading fixture3 should load fixture1 and fixture2 also
        fm.get_fixture('fixture3')
        self.assertIn('fixture1', fm.cache)
        self.assertIn('fixture4', fm.cache)

    def test_invalid_hook(self):
        """Verify that can't set an invalid hook."""
        manager = FixturesManager()
        with pytest.raises(KeyError):
            manager.set_hook("invalid", lambda p: p)

    def test_set_hook(self):
        """Verify that we can set a hook."""
        manager = FixturesManager()
        manager.set_hook("before_save", lambda p: p)



from __future__ import absolute_import

from charlatan import testcase, testing, FixturesManager


class TestRelationshipsWithoutModels(testing.TestCase,
                                     testcase.FixturesManagerMixin):

    def setUp(self):
        self.fixtures_manager = FixturesManager()
        self.fixtures_manager.load(
            './charlatan/tests/data/relationships_without_models.yaml')
        self.install_fixtures([
            'dict_with_nest', 'simple_dict', 'list_of_relationships'])
        self.init_fixtures()

    def test_dictionaries_nest(self):
        self.assertEqual(self.dict_with_nest['simple_dict'], self.simple_dict)

    def test_relationships_list(self):
        self.assertEqual([self.dict_with_nest, self.simple_dict],
                         self.list_of_relationships)

    def test_nested_list_of_relationships(self):
        nested_list_of_relationships = self.install_fixture(
            'nested_list_of_relationships')

        self.assertEqual(nested_list_of_relationships, {
            'dicts': [
                [self.dict_with_nest],
                [self.simple_dict],
            ]
        })

    def test_relationships_dict_attribute(self):
        parent = self.install_fixture('parent_dict.object1')
        child = self.install_fixture('child_dict.object1')

        self.assertEquals(child['field1'], parent['field1'])



from charlatan.utils import copy_docstring_from
from charlatan import FixturesManager
from charlatan.fixtures_manager import make_list


class FixturesManagerMixin(object):

    """Class from which test cases should inherit to use fixtures.

    .. versionchanged:: 0.3.12
        ``FixturesManagerMixin`` does not install class attributes
        ``fixtures`` anymore.

    .. versionchanged:: 0.3.0
        ``use_fixtures_manager`` method renamed ``init_fixtures.``

    .. versionchanged:: 0.3.0
        Extensive change to the function signatures.

    """

    def init_fixtures(self):
        """Initialize the fixtures.

        This function *must* be called before doing anything else.
        """
        self.fixtures_manager.clean_cache()

    @copy_docstring_from(FixturesManager)
    def install_fixture(self, fixture_key, overrides=None):
        fixture = self.fixtures_manager.install_fixture(fixture_key, overrides)
        setattr(self, fixture_key, fixture)
        return fixture

    @copy_docstring_from(FixturesManager)
    def install_fixtures(self, fixtures):
        installed = []
        for fixture in make_list(fixtures):
            installed.append(
                self.install_fixture(fixture)
            )

        return installed

    @copy_docstring_from(FixturesManager)
    def install_all_fixtures(self):
        return self.install_fixtures(self.fixtures_manager.keys())

    @copy_docstring_from(FixturesManager)
    def get_fixture(self, fixture_key, overrides=None):
        return self.fixtures_manager.get_fixture(fixture_key, overrides)

    @copy_docstring_from(FixturesManager)
    def get_fixtures(self, fixtures):
        return self.fixtures_manager.get_fixtures(fixtures)

    @copy_docstring_from(FixturesManager)
    def uninstall_fixture(self, fixture_key):
        self.fixtures_manager.uninstall_fixture(fixture_key)

    @copy_docstring_from(FixturesManager)
    def uninstall_fixtures(self, fixtures):
        for fixture in make_list(fixtures):
            self.uninstall_fixture(fixture)

    @copy_docstring_from(FixturesManager)
    def uninstall_all_fixtures(self):
        # copy and reverse the list in order to remove objects with
        # relationships first
        installed_fixtures = list(self.fixtures_manager.installed_keys)
        installed_fixtures.reverse()
        return self.uninstall_fixtures(installed_fixtures)
